from __future__ import annotations
from collections import defaultdict
from typing import List
from shrillecho.api.api_client import SpotifyClient
from shrillecho.utility.track_utils import TrackUtils
from shrillecho.types.album_types import Album
from shrillecho.types.artist_types import Artist
from shrillecho.types.playlist_types import Playlist
from shrillecho.types.soundcloud_types import User
from shrillecho.types.track_types import Track
from shrillecho.utility.artist_utils import ArtistUtils


class TracksController:
    """
    A class to encapsulate any collection of tracks
    """

    def __init__(self, sp: SpotifyClient, tracks: List[Track] = []):
        self.tracks: List[Track] = tracks
        self.liked_tracks: List[Track] = []
        self.unliked_tracks: List[Track] = []
        self.loaded_likes = False
        self._sp = sp

    @classmethod
    async def create_from_playlist(cls, sp: SpotifyClient, playlist: str):
        return cls(
            sp,
            tracks=await sp.playlists.tracks_batch(playlist_id=playlist, batch_size=5),
        )

    @classmethod
    async def create_from_album(cls, sp: SpotifyClient, album: str):
        return cls(sp, tracks=await sp.albums.album_tracks_batch(album_id=album))

    async def get_likes(self) -> TracksController:
        if not self.loaded_likes:
            await self.load_likes()
        return TracksController(sp=self._sp, tracks=self.liked_tracks)

    async def get_unliked(self) -> TracksController:
        if not self.loaded_likes:
            await self.load_likes()
        return TracksController(sp=self._sp, tracks=self.unliked_tracks)

    async def load_playlist(self, playlist_id: str):
        self.tracks = await self._sp.playlists.tracks_batch(playlist_id=playlist_id)

    async def load_album(self, album_id: str):
        self.tracks = await self._sp.albums.album_tracks_batch(album=album_id)

    async def load_likes(self):
        saved_tracks: List[Track] = await self._sp.tracks.saved_tracks_batch()

        saved_track_map = set(saved_tracks)

        for track in self.tracks:
            if not track:
                print(track)

            if track in saved_track_map:
                track.liked = True
                self.liked_tracks.append(track)
            else:
                track.liked = False
                self.unliked_tracks.append(track)

        self.loaded_likes = True

    async def clean_tracks(self):
        """
        If a track has no ISRC then we presume its an invalid track
        """
        local_removed_copyright = 0
        cleaned_tracks: List[Track] = []
        for track in self.tracks:
            if track.external_ids.isrc:
                cleaned_tracks.append(track)
            else:
                local_removed_copyright += 1
        self.tracks = cleaned_tracks

    async def write_tracks(self, name: str, user: str = "me") -> Playlist:

        if user == "me":
            me: User = await self._sp.users.me()
            user = me.id

        track_ids = TrackUtils.fetch_track_ids(self.tracks)

        track_uris = [f'spotify:track:{id}' for id in track_ids]

        playlist = await self._sp.playlists.create_playlist(
            user_id=user, playlist_name=name, public=False
        )
        limit = 50
        for i in range(0, len(track_uris), limit):
            await self._sp.playlists.add_tracks(
                playlist_id=playlist.id, track_uris=track_uris[i : i + limit]
            )
        return playlist

    async def artists(self, all_artists=False) -> List[Artist]:
        artist_ids_to_fetch = set()
        for track in self.tracks:
            if all_artists:
                for artist in track.artists:
                    artist_ids_to_fetch.add(artist.id)
            else:
                main_artist: Artist = TrackUtils.determine_main_artist(track)
                if main_artist.id:
                    artist_ids_to_fetch.add(track.artists[0].id)

        playlist_artists: list[Artist] = await ArtistUtils.fetch_several_artists(
            self._sp, artist_ids=list(artist_ids_to_fetch)
        )
        return playlist_artists

    async def genres(self) -> List[bool]:
        """
        From our list of tracks, obtain all the genres in a dictionary form

        genres = {
            'future_bass': 5,
            'hyperpop': 2,
            ...
        }
        """
        playlist_artists: List[Artist] = await self.artists()

        genre_counts = defaultdict(int)

        for artist in playlist_artists:
            for genre in artist.genres:
                genre_counts[genre] += 1

        sorted_genres: List[int] = sorted(
            genre_counts.items(), key=lambda x: x[1], reverse=True
        )

        return sorted_genres

    def __sub__(self, other: TracksController):
        other_track_sets = set(other.tracks)
        return TracksController(
            sp=self._sp,
            tracks=[track for track in self.tracks if track not in other_track_sets],
        )

    def __repr__(self):
        return f"{self.tracks}"

    def __and__(self, other):
        if isinstance(other, TracksController):
            self_set = set(self.tracks)
            other_set = set(other.tracks)
            return TracksController(sp=self._sp, tracks=self_set & other_set)
        return NotImplemented

    def __getitem__(self, index) -> Track:
        return self.tracks[index]

    def __len__(self):
        return len(self.tracks)



from abc import ABC
import math
from typing import Dict, List
from shrillecho.api.api_request_handler import SpotifyApiRequestHandler
from shrillecho.api.task_scheduler import TaskScheduler


class Endpoint(ABC):
    def __init__(self, api_request_handler):
        self._request_handler: SpotifyApiRequestHandler = api_request_handler
        self._task_scheduler: TaskScheduler = TaskScheduler()

    @staticmethod
    def _format_url_params(url_template: str, map: dict):
        url = url_template
        for key, value in map.items():
            placeholder = f"{{{key}}}"
            url = url.replace(placeholder, str(value))
        return url

    @staticmethod
    def _generate_pagination_batch_urls(initial_item) -> List[Dict[str, int]]:

        if not hasattr(initial_item, "total"):
            raise AttributeError("initial_item must have a .total attribute")

        query_param_dicts: List[dict] = []
        limit = 50
        pages = math.ceil(initial_item.total / limit)
        for page in range(1, pages):
            query_param_dicts.append({"limit": 50, "offset": limit * page})
        return query_param_dicts

    async def batch_get(
        self, url: str, response_type, initial_item, batch_size, **kwargs
    ):

        query_param_offsets = self._generate_pagination_batch_urls(
            initial_item=initial_item
        )

        tasks = [
            (
                self._request_handler._req_proxy,
                ("GET", url, response_type, {**query_params, **kwargs}),
            )
            for query_params in query_param_offsets
        ]

        self._task_scheduler.load_tasks(tasks)

        return await self._task_scheduler.execute_tasks(
            batch_size=batch_size, unpack=True
        )



from typing import List
from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.types.album_types import (
    Album,
    AlbumTracks,
    SeveralAlbums,
    SimplifiedTrack,
)
from shrillecho.utility.general_utility import get_id
from .endpoint_urls import SPOTIFY_URLS


class AlbumsEndpoint(Endpoint):

    async def album_tracks_batcher(
        self, album_id: str, batch_size=1, **kwargs
    ) -> List[SimplifiedTrack]:

        response_type = AlbumTracks
        url = SPOTIFY_URLS.ALBUM.TRACKS.format(id=album_id)

        initial_query_params = {"offset": 0, "limit": 50}

        initial_item: AlbumTracks = await self._request_handler._req_proxy(
            method="GET",
            url=url,
            query_params={**initial_query_params, **kwargs},
            response_type=response_type,
        )

        album_tracks_response: List[SimplifiedTrack] = [t for t in initial_item.items]

        album_tracks_response.extend(
            await self.batch_get(
                url=url,
                response_type=response_type,
                initial_item=initial_item,
                batch_size=batch_size,
                **kwargs
            )
        )
        return album_tracks_response

    async def album_tracks_batch(
        self, album_id: str, batch_size: int = 1, **kwargs
    ) -> List[SimplifiedTrack]:
        return await self.album_tracks_batcher(
            batch_size=batch_size, album_id=get_id("album", album_id), **kwargs
        )

    async def one(self, album_id: str, **kwargs) -> Album:

        url = SPOTIFY_URLS.ALBUM.ONE.format(id=get_id("album", album_id))

        return await self._request_handler._req_proxy(
            method="GET", url=url, query_params={**kwargs}, response_type=Album
        )

    async def many(self, album_uris: List[str], **kwargs) -> List[Album]:

        url = SPOTIFY_URLS.ALBUM.MULTIPLE

        multiple_albums: SeveralAlbums = await self._request_handler.make_request(
            method="GET",
            url=url,
            query_params={"ids": album_uris, **kwargs},
            response_type=SeveralAlbums,
        )
        return multiple_albums.albums



from dataclasses import dataclass
from typing import List
from fastclasses_json import dataclass_json

from shrillecho.types.album_types import Album, SimplifiedAlbum
from shrillecho.types.artist_types import Artist
from shrillecho.types.playlist_types import SimplifiedPlaylistObject
from shrillecho.types.track_types import Track


@dataclass_json
@dataclass
class TrackSearchItems:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[Track]


@dataclass_json
@dataclass
class AlbumSearchItems:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[SimplifiedAlbum]


@dataclass_json
@dataclass
class ArtistSearchItems:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[Artist]


@dataclass_json
@dataclass
class PlaylistSearchItems:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[SimplifiedPlaylistObject]


@dataclass_json
@dataclass
class SearchResult:
    tracks: TrackSearchItems
    artists: ArtistSearchItems
    albums: AlbumSearchItems
    playlists: PlaylistSearchItems



import asyncio
from collections import deque
from dataclasses import asdict
import json
from typing import List, Set, Tuple

from pymongo import MongoClient
from shrillecho.api.api_client import SpotifyClient
from shrillecho.controllers.artist_controller import ArtistController
from shrillecho.types.artist_types import Artist
from shrillecho.utility.artist_utils import ArtistUtils
from shrillecho.utility.cache import Mongo
from shrillecho.utility.neo4j import NeoController


class Mongo:
    def __init__(self, domain: str, port: str, db: str = None):
        self._mongo_client = MongoClient(f"mongodb://{domain}:{port}/")
        self._db = self._mongo_client[db]

    def insert_document(self, collection_name: str, document: dict):
        collection = self._db[collection_name]
        collection.insert_one(document)

    def initialise_artist_state(self, artist: Artist) -> List[Artist]:
        collection = self._db["graph_state"]
        document = collection.find_one({"_id": artist.id})
        if not document:

            new_artist_state = {"_id": artist.id, "exp": [], "unexp": [asdict(artist)]}

            if collection.insert_one(new_artist_state):
                print("inserted new artist state okay")
            else:
                print("insertion failure")
                exit(1)

    def update_array(self, artist: Artist, array_type: str, new_array) -> List[Artist]:
        collection = self._db["graph_state"]
        collection.update_one(
            {"_id": artist.id},
            {"$set": {array_type: [asdict(item) for item in new_array]}},
        )

    def pop_unexplored(self, artist: Artist) -> Artist:
        collection = self._db["graph_state"]

        unexplored = collection.find_one_and_update(
            {"_id": artist.id}, {"$pop": {"unexp": -1}}, return_document=True
        )

        return Artist.from_json(json.dumps(unexplored))

    def get_explored_states(self, artist: Artist):
        collection = self._db["graph_state"]

        state = collection.find_one({"_id": artist.id})
        if state:

            return {
                "exp": [Artist.from_json(json.dumps(a)) for a in state["exp"]],
                "unexp": [Artist.from_json(json.dumps(a)) for a in state["unexp"]],
            }
        else:

            data = {"_id": artist.id, "exp": [], "unexp": [asdict(artist)]}

            collection.insert_one(data)

            return {
                "exp": [],
                "unexp": [Artist.from_json(json.dumps(a)) for a in data["unexp"]],
            }


class ArtistCrawler:

    def __init__(self, sp: SpotifyClient):
        self._sp = sp
        self._neo = NeoController(username="neo4j", password="Zxasqw12")
        self._mongo = Mongo("localhost", "27017", "shrillecho")

    # async def bfs_related(self, start: Artist, max_depth: int = 2) -> List[Artist]:

    #     unexpanded, expanded = self._mongo.get_graph_state()
    #     # unexpanded: deque[Tuple[Artist, int]] = deque([(start, 0)])

    #     while unexpanded:

    #         cur_artist, depth = unexpanded.popleft()

    #         if cur_artist.id not in expanded and depth <= max_depth:

    #             related_artists: List[Artist] = await self._sp.artists.related(artist_id=cur_artist.id)
    #             expanded.add(cur_artist.id)

    #             for artist in related_artists:
    #                 if depth + 1 < max_depth and artist.id not in expanded:
    #                     unexpanded.append((artist, depth + 1))

    #     return await ArtistUtils.fetch_several_artists(sp=self._sp,artist_ids=list(expanded))

    def get_full_list(self, expanded, unexpanded) -> List[Artist]:
        expanded_list = list(expanded)
        expanded_list.extend(list(set(unexpanded)))
        return expanded_list

    async def related(self, artist: str) -> List[Artist]:
        return await self._sp.artists.related(artist_id=artist)

    async def bfs_related_state(
        self, start: Artist, cursor: Artist = 0, limit: int = 50
    ) -> List[Artist]:
        graph_state = self._mongo.get_explored_states(artist=start)

        expanded = set(graph_state["exp"])
        unexpanded = deque(graph_state["unexp"])

        while True:

            batch: List[Artist] = []

            for _ in range(0, min(len(unexpanded), 50)):
                cur_artist: Artist = unexpanded.popleft()
                batch.append(cur_artist)

            if any(batch) not in expanded:

                related_artists: List[List[Artist]] = await asyncio.gather(
                    *[self.related(ba.id) for ba in batch]
                )
                flatten_related = [
                    item for sublist in related_artists for item in sublist
                ]

                for item in batch:
                    expanded.add(item)

                for artist in flatten_related:
                    if artist not in expanded:
                        unexpanded.append((artist))

            self._mongo.update_array(artist=start, array_type="exp", new_array=expanded)
            self._mongo.update_array(
                artist=start, array_type="unexp", new_array=unexpanded
            )



from typing import List
from shrillecho.api.api_client import SpotifyClient
from shrillecho.types.artist_types import Artist
from shrillecho.types.track_types import Track


class TrackUtils:

    @staticmethod
    async def fetch_several_tracks(
        sp: SpotifyClient, track_ids: List[str], batch_size: int = 50
    ) -> List[Track]:

        tracks: List[Track] = []
        for i in range(0, len(track_ids), batch_size):
            chunk = track_ids[i : i + batch_size]
            more_tracks: List[Track] = await sp.tracks.many(chunk)
            tracks.extend(more_tracks)

        return tracks

    @staticmethod
    def determine_main_artist(track: Track) -> Artist:
        """The main artist is the artist who actually made the track,
        so if its a remix , they are the creators of the remix but may not be listed
        as the first artist. The main artist does not include the feats just the core artist,
        may extend to be feats..

        Args:
            track (Track): the track to determine the main artist of

        Returns:
            List[Artist]: list of artists in the playlist
        """
        main_artist = None
        if "Remix" in track.name:
            for artist in track.artists:
                if artist.name in track.name:
                    main_artist = artist
            if not main_artist:
                main_artist = track.artists[0]
        else:
            main_artist = track.artists[0]

        return main_artist

    @staticmethod
    def fetch_track_ids(tracks: List[Track]) -> List[str]:
        """
        Given a list of tracks return a list of the ids only
        """

        ids = []
        for track in tracks:
            if track.id:
                ids.append(track.id)

        print(f"ids: {len(ids)}")
        return ids

    @staticmethod
    def clean_tracks(tracks: List[Track]):

        cleaned_tracks: List[Track] = []
        for track in tracks:
            if track.external_ids.isrc:
                cleaned_tracks.append(track)
        return cleaned_tracks



from typing import List
from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.api.endpoints.endpoint_urls import SPOTIFY_URLS
from shrillecho.types.artist_types import Artist
from shrillecho.types.player_types import PlaybackState, RecentlyPlayedTracks


class PlayerEndpoint(Endpoint):

    async def recently_played(
        self, after: int = None, before: int = None, limit: int = 50
    ) -> RecentlyPlayedTracks:
        """TODO: complete"""

        url = SPOTIFY_URLS.PLAYER.RECENTLY

        query_params = {"limit": limit}

        if after:
            query_params["after"] = after
            print("after")

        elif before:
            query_params["before"] = before
            print("before")

        # else:
        #     raise Exception("invalid call")

        return await self._request_handler.make_request(
            method="GET",
            url=url,
            query_params=query_params,
            response_type=RecentlyPlayedTracks,
        )

    async def playback_state(self, **kwargs) -> PlaybackState:
        """TODO: complete"""

        url = SPOTIFY_URLS.PLAYER.PLAYER

        return await self._request_handler.make_request(
            method="GET", url=url, query_params=kwargs, response_type=PlaybackState
        )



from typing import List
from shrillecho.api.api_client import SpotifyClient
from shrillecho.types.album_types import Album, SeveralAlbums


class AlbumUtils:

    @staticmethod
    async def fetch_several_albums(
        sp: SpotifyClient, album_ids: List[str], batch_size: int = 20
    ) -> List[Album]:
        albums: List[Album] = []
        for i in range(0, len(album_ids), batch_size):
            chunk = album_ids[i : i + batch_size]
            more_albums: SeveralAlbums = await sp.albums(chunk)
            albums.extend(more_albums.albums)
        return albums



from typing import List, Set
from shrillecho.api.api_client import SpotifyClient
from shrillecho.controllers.tracks_controller import TracksController
from shrillecho.types.track_types import Track
from shrillecho.utility.general_utility import get_id


class TrackController:

    def __init__(self, sp: SpotifyClient, track: Track = None):
        self._sp = sp
        self._track = track
        self._liked_tracks: Set[Track] = None
        self._loaded_likes = False

    @classmethod
    def create(cls, track_uri: str):
        return cls(track=get_id("track", track_uri))

    async def _load_likes(self):
        saved_tracks: List[Track] = await self._sp.tracks.saved_tracks_batch()
        self._liked_tracks = set(saved_tracks)
        self._loaded_likes = True

    async def is_liked(self) -> bool:
        if not self._loaded_likes:
            await self._load_likes()
        return self._track in self._liked_tracks



from typing import List
from shrillecho.controllers.tracks_controller import TracksController
from shrillecho.api.api_client import SpotifyClient
from shrillecho.types.playlist_types import Playlist
from shrillecho.types.track_types import Track


class PlaylistController:

    def __init__(
        self,
        sp: SpotifyClient,
        playlist: Playlist = None,
        tracks: TracksController = None,
    ):

        if not playlist or not tracks:
            raise Exception(
                "Use `create` methods or pass track controller and playlist"
            )

        self._sp: SpotifyClient = sp
        self.playlist: Playlist = playlist
        self.playlist_tracks: TracksController = tracks

    @classmethod
    async def create(
        cls, sp: SpotifyClient, playlist_id: str = None, playlist: Playlist = None
    ):

        if not playlist:
            playlist = await sp.playlist(playlist_id=playlist_id)

        if not playlist_id:
            playlist_id = playlist.id

        tracks = await TracksController.create_from_playlist(sp, playlist=playlist_id)

        return cls(sp=sp, playlist=playlist, tracks=tracks)

    def internal_tracks(self) -> List[Track]:
        return self.playlist_tracks.tracks



from typing import List
import spotipy

from shrillecho.api.api_client import SpotifyClient
from shrillecho.types.track_types import Track


class SpotifyPlayer:

    @staticmethod
    def play_track(sp: spotipy.Spotify, uri: str):
        """
        Play a specific uri on the first device ID listed.
        """

        uri = f"spotify:track:{uri}"

        device_id = sp.devices()["devices"][0]["id"]
        sp.start_playback(device_id=device_id, uris=[uri])

    @staticmethod
    async def get_recently_played(sp: SpotifyClient) -> List[Track]:
        sp.player.current_user_recently_played()



from typing import Union

from shrillecho.api.endpoints.search import SearchEndpoint

from .endpoints.albums import AlbumsEndpoint
from .endpoints.artists import ArtistsEndpoint
from .endpoints.follow import FollowEndpoint
from .endpoints.player import PlayerEndpoint
from .endpoints.playlists import PlaylistEndpoint
from .endpoints.tracks import TracksEndpoint
from .endpoints.users import UsersEndpoint
from .auth.oauth import ClientCredentials, OAuthCredentials
from .api_request_handler import SpotifyApiRequestHandler


class SpotifyClient:

    def __init__(
        self, token=None, auth_flow: Union[OAuthCredentials | ClientCredentials] = None
    ):

        if auth_flow:
            self.access_token = auth_flow.get_access_token()
        else:
            self.access_token = token

        self._request_handler = SpotifyApiRequestHandler(token=self.access_token)

        self.playlists: PlaylistEndpoint = PlaylistEndpoint(self._request_handler)
        self.tracks: TracksEndpoint = TracksEndpoint(self._request_handler)
        self.albums: AlbumsEndpoint = AlbumsEndpoint(self._request_handler)
        self.artists: ArtistsEndpoint = ArtistsEndpoint(self._request_handler)
        self.users: UsersEndpoint = UsersEndpoint(self._request_handler)
        self.follow: FollowEndpoint = FollowEndpoint(self._request_handler)
        self.player: PlayerEndpoint = PlayerEndpoint(self._request_handler)
        self.search: SearchEndpoint = SearchEndpoint(self._request_handler)



from typing import List
from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.api.endpoints.endpoint_urls import SPOTIFY_URLS
from shrillecho.types.artist_types import Artist
from shrillecho.types.user_types import CurrentUserProfile
from shrillecho.utility.general_utility import get_id


class FollowEndpoint(Endpoint):

    async def get_followed_artists(self, after: str, limit: int) -> List[Artist]:

        url = SPOTIFY_URLS.FOLLOW.FOLLOWING_ARTISTS

        return await self._request_handler.make_request(
            method="GET",
            url=url,
            query_params={"after": after, "limit": limit},
            response_type=CurrentUserProfile,
        )

    async def current_user_unfollow_playlist(self, playlist_id: str):

        url = SPOTIFY_URLS.FOLLOW.PLAYLIST.format(
            playlist_id=get_id("playlist", playlist_id)
        )

        await self._request_handler.make_request(
            method="DELETE", url=url, query_params={}
        )



from typing import List
from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.types.album_types import ArtistAlbums, SimplifiedAlbum
from shrillecho.types.artist_types import Artist, SeveralArtists
from shrillecho.types.track_types import SavedTracks
from shrillecho.utility.general_utility import get_id
from .endpoint_urls import SPOTIFY_URLS


class ArtistsEndpoint(Endpoint):

    async def artist_albums_batcher(
        self, artist_id: str, batch_size=1, **kwargs
    ) -> List[SimplifiedAlbum]:

        response_type = ArtistAlbums
        url = SPOTIFY_URLS.ARTIST.ALBUM.format(id=artist_id)

        initial_query_params = {"offset": 0, "limit": 50}

        initial_item: ArtistAlbums = await self._request_handler._req_proxy(
            method="GET",
            url=url,
            query_params={**initial_query_params, **kwargs},
            response_type=response_type,
        )

        artist_albums_response: List[SimplifiedAlbum] = initial_item.items

        artist_albums_response.extend(
            await self.batch_get(
                url=url,
                response_type=response_type,
                initial_item=initial_item,
                batch_size=batch_size,
                **kwargs
            )
        )
        return artist_albums_response

    async def artist_albums_batch(
        self, artist_id: str, batch_size: int = 1, **kwargs
    ) -> List[SimplifiedAlbum]:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        return await self.artist_albums_batcher(
            batch_size=batch_size, artist_id=get_id("artist", artist_id), **kwargs
        )

    async def artist_albums(self, artist_id: str, **kwargs) -> List[SimplifiedAlbum]:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        url = SPOTIFY_URLS.ARTIST.ALBUM.format(id=artist_id)

        return await self._request_handler.make_request(
            method="GET",
            url=url,
            query_params={**kwargs},
            response_type=SimplifiedAlbum,
        )

    async def get(self, artist_id: str) -> Artist:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        url = SPOTIFY_URLS.ARTIST.ONE.format(id=get_id("artist", artist_id))

        return await self._request_handler._req_proxy(
            method="GET", url=url, query_params={}, response_type=Artist
        )

    async def many(self, artist_ids: List[str], **kwargs) -> List[Artist]:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        url = SPOTIFY_URLS.ARTIST.SEVERAL

        artists: SeveralArtists = await self._request_handler._req_proxy(
            method="GET",
            url=url,
            query_params={**{"ids": artist_ids}, **kwargs},
            response_type=SeveralArtists,
        )

        return artists

    async def related(self, artist_id: str) -> List[Artist]:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        url = SPOTIFY_URLS.ARTIST.SIMILAR_ARTISTS.format(id=get_id("artist", artist_id))

        related_artists: SeveralArtists = await self._request_handler._req_proxy(
            method="GET", url=url, query_params={}, response_type=SeveralArtists
        )
        return related_artists.artists



from typing import List
from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.types.playlist_types import (
    Playlist,
    PlaylistTracks,
    SimplifiedPlaylistObject,
    UserPlaylists,
)
from shrillecho.types.track_types import Track
from shrillecho.utility.general_utility import get_id
from .endpoint_urls import SPOTIFY_URLS


class PlaylistEndpoint(Endpoint):

    async def playlist_tracks_batcher(
        self, playlist_id: str, batch_size=1, **kwargs
    ) -> List[Track]:

        response_type = PlaylistTracks

        url = SPOTIFY_URLS.PLAYLIST.TRACKS.format(playlist_id=playlist_id)

        initial_query_params = {"offset": 0, "limit": 50}

        initial_item: PlaylistTracks = await self._request_handler._req_proxy(
            method="GET",
            url=url,
            query_params={**initial_query_params, **kwargs},
            response_type=response_type,
        )

        playlist_tracks_response: List[Track] = [t.track for t in initial_item.items]

        playlist_tracks_response.extend(
            await self.batch_get(
                url=url,
                response_type=response_type,
                initial_item=initial_item,
                batch_size=batch_size,
                **kwargs
            )
        )

        return playlist_tracks_response

    async def current_user_playlists_batcher(
        self, batch_size=1, **kwargs
    ) -> List[Track]:

        response_type = UserPlaylists

        url = SPOTIFY_URLS.PLAYLIST.MY_PLAYLISTS

        initial_query_params = {"offset": 0, "limit": 50}

        initial_item: UserPlaylists = await self._request_handler._req_proxy(
            method="GET",
            url=url,
            query_params={**initial_query_params, **kwargs},
            response_type=response_type,
        )

        user_playlists_response: List[SimplifiedPlaylistObject] = initial_item.items

        user_playlists_response.extend(
            await self.batch_get(
                url=url,
                response_type=response_type,
                initial_item=initial_item,
                batch_size=batch_size,
                **kwargs
            )
        )

        return user_playlists_response

    async def user_playlists_batcher(
        self, user_id: str, batch_size=1, **kwargs
    ) -> List[Track]:

        response_type = UserPlaylists

        url = SPOTIFY_URLS.PLAYLIST.USER.format(user_id=user_id)

        initial_query_params = {"offset": 0, "limit": 50}

        initial_item: UserPlaylists = await self._request_handler.make_request(
            method="GET",
            url=url,
            query_params={**initial_query_params, **kwargs},
            response_type=response_type,
        )

        user_playlists_response: List[SimplifiedPlaylistObject] = initial_item.items

        user_playlists_response.extend(
            await self.batch_get(
                url=url,
                response_type=response_type,
                initial_item=initial_item,
                batch_size=batch_size,
                **kwargs
            )
        )

        return user_playlists_response

    async def create_playlist(
        self, user_id: str, playlist_name: str, **kwargs
    ) -> Playlist:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/create-playlist
        """

        url = SPOTIFY_URLS.PLAYLIST.CREATE.format(user_id=get_id("user", user_id))

        return await self._request_handler.make_request(
            method="POST",
            url=url,
            query_params={},
            response_type=Playlist,
            body={**{"name": playlist_name}, **kwargs},
            ignore_cache=True
        )

    async def get(self, playlist_id: str, **kwargs) -> Playlist:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlist
        """

        url = SPOTIFY_URLS.PLAYLIST.ONE.format(
            playlist_id=get_id("playlist", playlist_id)
        )

        return await self._request_handler.make_request(
            method="GET", url=url, query_params=kwargs, response_type=Playlist
        )

    async def tracks_batch(
        self, playlist_id: str, batch_size: int = 50, **kwargs
    ) -> Playlist:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        return await self.playlist_tracks_batcher(
            playlist_id=get_id("playlist", playlist_id), batch_size=batch_size, **kwargs
        )

    async def current_user_playlists_batch(
        self, batch_size: int = 1, **kwargs
    ) -> Playlist:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        return await self.current_user_playlists_batcher(
            batch_size=batch_size, **kwargs
        )

    async def user_playlists_batch(
        self, user_id: str, batch_size: int = 1, **kwargs
    ) -> Playlist:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        return await self.user_playlists_batcher(
            user_id=get_id("user", user_id), batch_size=batch_size, **kwargs
        )

    async def add_tracks(
        self, playlist_id: str, track_uris: List[str], position: int = None
    ) -> str:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        url = SPOTIFY_URLS.PLAYLIST.TRACKS.format(playlist_id=get_id("playlist", playlist_id))

        body = {"uris": track_uris}
    

        if position is not None:
            body["position"] = position

        return await self._request_handler.make_request(
            method="POST", url=url, query_params={}, body=body, ignore_cache=True
        )



from typing import List
from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.types.playlist_types import Playlist, PlaylistTracks
from shrillecho.types.track_types import (
    AudioFeatures,
    Recc,
    SavedTracks,
    SeveralTracks,
    Track,
)
from shrillecho.utility.general_utility import get_id
from .endpoint_urls import SPOTIFY_URLS


class TracksEndpoint(Endpoint):

    async def _saved_tracks_batcher(self, batch_size=1, **kwargs) -> List[Track]:

        response_type = SavedTracks
        url = SPOTIFY_URLS.LIBRARY.TRACKS

        initial_query_params = {"offset": 0, "limit": 50}

        # We need to know the total number of tracks to collect all URLs, so we do this initial call
        initial_item: SavedTracks = await self._request_handler.make_request(
            method="GET",
            url=url,
            query_params={**initial_query_params, **kwargs},
            response_type=response_type,
        )

        # unpack the initial response to our base response
        saved_tracks_response: List[Track] = [t.track for t in initial_item.items]

        # extend the base response with a batched get of all other urls
        saved_tracks_response.extend(
            await self.batch_get(
                url=url,
                response_type=response_type,
                initial_item=initial_item,
                batch_size=batch_size,
                **kwargs,
            )
        )
        return saved_tracks_response

    async def saved_tracks_batch(self, batch_size: int = 1, **kwargs) -> List[Track]:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        return await self._saved_tracks_batcher(batch_size=batch_size, **kwargs)

    async def audio_features(self, track_id: str) -> AudioFeatures:
        """
        docs: https://developer.spotify.com/documentation/web-api/reference/get-playlists-tracks
        """

        url = SPOTIFY_URLS.TRACKS.FEATURES.format(id=get_id("track", track_id))

        return await self._request_handler._req_proxy(
            method="GET", endpoint=url, response_type=AudioFeatures
        )

    async def one(self, track_id: str, **kwargs) -> Track:

        url = SPOTIFY_URLS.TRACKS.ONE.format(id=get_id("track", track_id))

        return await self._request_handler._req_proxy(
            method="GET", url=url, query_params={**kwargs}, response_type=Track
        )

    async def many(self, track_ids: List[str], **kwargs) -> List[Track]:

        url = SPOTIFY_URLS.TRACKS.SEVERAL

        several_tracks: SeveralTracks = await self._request_handler._req_proxy(
            method="GET",
            url=url,
            query_params={**{"ids": track_ids}, **kwargs},
            response_type=SeveralTracks,
        )
        print(f"many: {len(several_tracks.tracks)}")
        return several_tracks.tracks

    async def reccomendation(self, **kwargs) -> Recc:

        url = SPOTIFY_URLS.TRACKS.RECOMMENDATIONS

        return await self._request_handler.make_request(
            method="GET", url=url, query_params={**kwargs}, response_type=Recc
        )



from typing import List
from shrillecho.controllers.tracks_controller import TracksController
from shrillecho.api.archive.client import SpotifyClient
from shrillecho.types.album_types import Album
from shrillecho.types.track_types import Track


class AlbumController:

    def __init__(self, sp: SpotifyClient, album: Album, tracks: TracksController):

        self._sp: SpotifyClient = sp
        self.album: Album = album
        self.album_tracks: TracksController = tracks

    @staticmethod
    async def create(sp: SpotifyClient, album_id: str = None, album: Album = None):

        if not album:
            album = await sp.album(album_id=album_id)

        if not album_id:
            album_id = album.id

        tracks = await TracksController.create_from_album(album_id)

        return AlbumController(sp=sp, album=album, tracks=tracks)

    async def internal_tracks(self) -> List[Track]:
        return self.album_tracks.tracks



from typing import List
from neo4j import GraphDatabase

from shrillecho.types.artist_types import Artist

# uri = "bolt://localhost:7687"
# username = "neo4j"
# password = "Zxasqw12"

# driver = GraphDatabase.driver(uri, auth=(username, password))

# def create_and_relate_artist(main_artist, other_artists):

#     def create_and_relate_artists(tx, main_artist: Artist, other_artists: List[Artist]):

#         main_artist_query = "MERGE (main:Artist {name: $main_artist}) RETURN main"
#         tx.run(main_artist_query, main_artist=main_artist.name)

#         for artist in other_artists:
#             # For each artist in the list, create or find the artist node and create a RELATED_TO relationship
#             artist_query = """
#             MERGE (other:Artist {name: $artist})
#             WITH other
#             MATCH (main:Artist {name: $main_artist})
#             MERGE (main)-[:RELATED_TO]->(other)
#             """
#             tx.run(artist_query, main_artist=main_artist.name, artist=artist.name)


#     with driver.session() as session:
#         session.execute_write(create_and_relate_artists, main_artist, other_artists)


class NeoController:

    def __init__(
        self,
        uri: str = "bolt://localhost:7687",
        username: str = None,
        password: str = None,
    ):
        if not (username and password):
            raise Exception("Provide authentication details!")

        self.__driver = GraphDatabase.driver(uri, auth=(username, password))

    def create_artist(self, tx, artist_name):
        query = "MERGE (a:Artist {name: $artist_name}) " "RETURN a"
        tx.run(query, artist_name=artist_name)

    def create_relationship(self, tx, artist_name, related_artist_name):
        query = (
            "MATCH (a:Artist {name: $artist_name}) "
            "MATCH (b:Artist {name: $related_artist_name}) "
            "MERGE (a)-[:RELATED_TO]->(b)"
        )
        tx.run(query, artist_name=artist_name, related_artist_name=related_artist_name)

    def set_unrelated_artist(self, tx):
        query = (
            "MATCH (a:Artist) "
            "WHERE NOT (a)-[:RELATED_TO]->() "
            "REMOVE a:Artist "
            "SET a:UnrelatedArtist, a.color = 'red'"
        )
        tx.run(query)

    def add_artists_with_relationships(self, artist_data):
        with self.__driver.session() as session:
            for artist in artist_data:
                artist_name = artist["name"]
                related_artists = artist["related"]
                session.write_transaction(self.create_artist, artist_name)
                for related_artist in related_artists:
                    session.write_transaction(self.create_artist, related_artist)
                    session.write_transaction(
                        self.create_relationship, artist_name, related_artist
                    )
            # Set unrelated artists after all nodes and relationships are created
            session.write_transaction(self.set_unrelated_artist)



from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.api.endpoints.endpoint_urls import SPOTIFY_URLS
from shrillecho.types.user_types import CurrentUserProfile


class UsersEndpoint(Endpoint):

    async def me(self) -> CurrentUserProfile:

        url = SPOTIFY_URLS.USER.ME

        return await self._request_handler.make_request(
            method="GET", url=url, query_params={}, response_type=CurrentUserProfile
        )



from typing import List
from shrillecho.api.api_client import SpotifyClient
from shrillecho.types.artist_types import Artist, SeveralArtists
from shrillecho.types.track_types import Track


class ArtistUtils:

    @staticmethod
    async def fetch_several_artists(
        sp: SpotifyClient, artist_ids: List[str], batch_size: int = 20
    ) -> List[Artist]:
        artists: List[Artist] = []
        for i in range(0, len(artist_ids), batch_size):
            chunk = artist_ids[i : i + batch_size]
            more_artists: SeveralArtists = await sp.artists.many(artist_ids=chunk)
            artists.extend(more_artists.artists)
        return artists

    @staticmethod
    async def expand_tracks_list_with_artists(
        sp: SpotifyClient, tracks: List[Track], batch_size: int = 20
    ) -> List[Track]:
        # Collect the artist IDs to expand
        artist_ids = [track.artists[0].id for track in tracks]

        expanded_artists: List[Artist] = []
        for i in range(0, len(artist_ids), batch_size):
            chunk = artist_ids[i : i + batch_size]
            more_artists: SeveralArtists = await sp.artists.many(artist_ids=chunk)
            expanded_artists.extend(more_artists.artists)

        artist_mapping = {artist.id: artist for artist in expanded_artists}

        for track in tracks:
            track.artists[0] = artist_mapping[track.artists[0].id]

        return tracks

    @staticmethod
    def fetch_artists_ids(artists: List[Artist]) -> List[str]:
        """
        Given a list of tracks return a list of the ids only
        """

        ids = []
        for a in artists:
            if a.id:
                ids.append(a.id)
        return ids

    @staticmethod
    def return_top_genres(artists: List[Artist]) -> List[str]:
        """
        Given a list of tracks return a list of the ids only
        """



import asyncio
import json
from typing import Dict, List, Optional, Tuple
import httpx
from .endpoints.endpoint_urls import BASE_API_URL
import redis
from .endpoints.endpoint_urls import *


class SpotifyApiRequestHandler:

    class SpotifyBaseError(Exception):
        """
        The base Error for all spotify exceptions
        """

        def __init__(self, message: dict):
            self.message: dict = message

        def __str__(self):
            return str(self.message)

        def get_json(self) -> Dict[str, Dict[str, str]]:
            """
            Get the the api response which was an error as dict
            {"error":{"status": 400, "message": "The reason"}}
            """
            return self.message

    class RateLimit(SpotifyBaseError):
        def __init__(self, message, retry_after: int = None):
            super().__init__(message)
            self.retry_after = retry_after

    class TokenExpired(SpotifyBaseError):
        """
        Custom token expired message
        """

    class NotFound(SpotifyBaseError):
        """
        Custom token expired message
        """

    class SpotifyApiError(SpotifyBaseError):
        """
        Custom token expired message
        """

    def __init__(self, token: str):
        self.http_client: httpx.AsyncClient = httpx.AsyncClient()
        self.rate_lock = asyncio.Event()
        self._token = token
        self._redis = redis.Redis(host="192.168.0.132", port="6379", db=0)

    @staticmethod
    def _format_query_params(query_params: dict) -> List[Tuple[str, str]]:
        """
        Takes in a dictionary of query params may have a list as the value
        we basically just want to combine that list with commas

        if there is no list then we just pass in the value as an empty list
        """
        return_params: List[Tuple[str, str]] = []
        for key in list(query_params.keys()):
            if not isinstance(query_params[key], List):
                query_params[key] = [str(query_params[key])]
            return_params.append((key, ",".join([str(i) for i in query_params[key]])))

        return return_params

    async def _get_cache_key(self, method: str, url: str, query_params: dict) -> str:
        params_str = json.dumps(self._format_query_params(query_params), sort_keys=True)
        return f"{method}:{url}:{params_str}"

    async def make_request(
        self,
        method: str,
        url: str,
        response_type=None,
        query_params: dict = None,
        body: dict = None,
        ignore_cache: bool = False
    ):

        cache_key = await self._get_cache_key(method, url, query_params)
        cached_response = self._redis.get(cache_key)

        if cached_response and not ignore_cache:
            if response_type:

                return response_type.from_json(cached_response)
            else:
                return json.loads(cached_response)

        while self.rate_lock.is_set():
            await asyncio.sleep(1)

        query_params = self._format_query_params(query_params)

        headers = {
            "Authorization": f"Bearer {self._token}",
            "Content-Type": "application/json",
        }

        print(method, url, body)

        response = await self.http_client.request(
            method, url, params=query_params, headers=headers, json=body
        )
        print(response.text)

        if response.status_code == 429:
            print("429")
            self.rate_lock.set()
            retry_after = int(response.headers.get("Retry-After"))
            print(response.headers.get("Retry-After"))
            await asyncio.sleep(retry_after)
            self.rate_lock.clear()
            raise SpotifyApiRequestHandler.RateLimit(
                retry_after=retry_after, message=response.text
            )

        if response.status_code == 401:
            raise SpotifyApiRequestHandler.TokenExpired(message=response.text)

        if response.status_code == 404:
            raise SpotifyApiRequestHandler.NotFound(message=response.text)

        if response.status_code > 300:
            raise SpotifyApiRequestHandler.SpotifyApiError(message=response.text)

        self._redis.set(cache_key, json.dumps(response.json()), ex=3600)

        if response_type:

            with open("debug.json", "w") as f:
                json.dump(response.json(), f)
            return response_type.from_json(json.dumps(response.json()))
        else:

            return response.json()

    async def _req_proxy(
        self,
        method: str,
        url: str,
        response_type=None,
        query_params: Optional[dict] = None,
        body: Optional[dict] = None,
    ):

        if True:
            return await self.make_request(
                method, url, response_type, query_params, body
            )

        cache_key = await self._get_cache_key(method, url, query_params)
        cached_response = self._redis.get(cache_key)

        if cached_response:
            if response_type:

                return response_type.from_json(cached_response)
            else:
                return json.loads(cached_response)

        while self.rate_lock.is_set():
            await asyncio.sleep(1)

        data = {"params": query_params, "body": body, "method": method, "endpoint": url}

        response = await self.http_client.post("http://localhost:8002/req", json=data)
        resp_json = response.json()

        data = resp_json["data"]
        status_code = resp_json["status_code"]
        retry_after = resp_json.get("retry_after")

        if status_code == 429:
            print("429 proxied")
            self.rate_lock.set()
            retry_after = int(retry_after)
            print(retry_after)
            await asyncio.sleep(retry_after)
            self.rate_lock.clear()
            raise SpotifyApiRequestHandler.RateLimit(
                retry_after=retry_after, message="Rate limit exceeded"
            )

        if status_code == 401:
            raise SpotifyApiRequestHandler.TokenExpired(message="Unauthorised")

        if status_code == 404:
            raise SpotifyApiRequestHandler.NotFound(message="Not found")

        if status_code > 300:
            raise SpotifyApiRequestHandler.SpotifyApiError(message="API error")

        self._redis.set(cache_key, json.dumps(data), ex=3600)

        if response_type:
            return response_type.from_json(json.dumps(data))
        else:
            return data



from abc import ABC, abstractmethod
import datetime
from functools import wraps
import hashlib
import json
from dataclasses import asdict
from pymongo import MongoClient
import redis


class Cache(ABC):

    @abstractmethod
    async def cache_query(
        self,
        cache_key: str,
        cache_update_function,
        expiry: int,
        class_type: type,
        **kwargs,
    ):
        pass


class Mongo(Cache):
    def __init__(self, domain: str, port: str, db: str = None):
        self._mongo_client = MongoClient(f"mongodb://{domain}:{port}/")
        self._db = self._mongo_client[db]

    def set_db(self, db: str):
        self._db = self._mongo_client[db]

    async def cache_query(
        self, cache_key, cache_update_function, expiry, class_type, **kwargs
    ):
        collection = self._db[kwargs.get("col")]

        doc = collection.find_one({"_id": cache_key})

        if doc:
            print(f"using cached: {cache_key}")

            doc.pop("expireAt", None)
            if "items" in doc and isinstance(doc["items"], list):
                return [class_type.from_json(json.dumps(item)) for item in doc["items"]]
            else:
                return class_type.from_json(json.dumps(doc))

        print(f"fetching new: {cache_key}")
        new_data = await cache_update_function()

        if isinstance(new_data, list):
            data_dict = {"_id": cache_key, "items": [asdict(item) for item in new_data]}
        else:
            data_dict = asdict(new_data)
            data_dict["_id"] = cache_key

        data_dict["expireAt"] = datetime.datetime.now() + datetime.timedelta(
            seconds=expiry
        )

        collection.replace_one({"_id": cache_key}, data_dict, upsert=True)

        return new_data


class Redis(Cache):

    def __init__(self, host: str, port: str):
        self.redis_client = redis.Redis(host=host, port=port, db=0)

    async def cache_query(
        self, cache_key: str, cache_update_function, expiry, class_type, **kwargs
    ):
        cached_value = self.redis_client.get(cache_key)
        if cached_value:
            retrieved_value = json.loads(cached_value)
            # print(f"[REDIS CACHE] Existing entry {cache_key}")
            if isinstance(retrieved_value, list):
                return [
                    class_type.from_json(json.dumps(item)) for item in retrieved_value
                ]

            return class_type.from_json(json.dumps(retrieved_value))
        else:
            # print(f"[REDIS CACHE] New entry {cache_key}")
            new_value = await cache_update_function()

            if isinstance(new_value, list):

                new_value_dict = []

                for item in new_value:
                    if item:
                        new_value_dict.append(asdict(item))
            else:

                print(type(new_value))
                new_value_dict = asdict(new_value)

            self.redis_client.set(cache_key, json.dumps(new_value_dict), ex=expiry)

            return new_value

    def invalidate_cache_by_prefix(self, prefix: str):
        pattern = f"{prefix}*"
        keys = self.redis_client.keys(pattern)
        if keys:
            self.redis_client.delete(*keys)


mongo_client = Mongo("localhost", "27017", "shrillecho")
redis_client: Redis = Redis("192.168.56.101", "6379")


def set_redis_host(host):
    print(f"updating redis to {host}")
    redis_client = Redis(host, "6379")


# Cache invalidation method
async def invalidate_cache(prefix: str):
    redis_client.invalidate_cache_by_prefix(prefix)


def cached(cache_key_prefix, class_type, expiry=3600):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):

            args_str = "_".join([str(arg) for arg in args[1:]])
            kwargs_str = "_".join([f"{k}_{v}" for k, v in kwargs.items()])
            dynamic_part = f"{args_str}_{kwargs_str}"

            dynamic_hash = hashlib.sha256(dynamic_part.encode()).hexdigest()

            full_cache_key = f"{cache_key_prefix}_{dynamic_hash}"

            async def update_function():
                return await func(*args, **kwargs)

            return await redis_client.cache_query(
                full_cache_key, update_function, expiry, class_type
            )

        return wrapper

    return decorator



from typing import List
from shrillecho.api.endpoints.endpoint import Endpoint
from shrillecho.api.endpoints.endpoint_urls import SPOTIFY_URLS
from shrillecho.types.artist_types import Artist
from shrillecho.types.search_types import SearchResult


class SearchEndpoint(Endpoint):

    async def search_item(
        self, query: str, query_type: List[str], **kwargs
    ) -> SearchResult:

        url = SPOTIFY_URLS.SEARCH.ITEM

        return await self._request_handler.make_request(
            method="GET",
            url=url,
            query_params={**{"q": query, "type": query_type}, **kwargs},
            response_type=SearchResult,
        )



from dataclasses import dataclass
from typing import List
from fastclasses_json import dataclass_json

from shrillecho.types.base_types import ExternalUrls
from shrillecho.types.track_types import Track


@dataclass_json
@dataclass
class Cursors:
    after: str
    before: str


@dataclass_json
@dataclass
class TrackHistoryContext:
    type: str
    href: str
    external_urls: ExternalUrls
    uri: str


@dataclass_json
@dataclass
class PlayHistoryObject:
    track: Track
    played_at: str
    context: TrackHistoryContext


@dataclass_json
@dataclass
class RecentlyPlayedTracks:
    """https://developer.spotify.com/documentation/web-api/reference/get-recently-played"""

    href: str
    limit: int
    next: str
    cursors: Cursors
    total: int
    items: List[PlayHistoryObject]


@dataclass_json
@dataclass
class Device:
    id: str
    is_active: bool
    is_private_session: bool
    is_restricted: bool
    name: str
    type: str
    volume_percent: int
    supports_volume: bool


@dataclass_json
@dataclass
class Context:
    type: str
    href: str
    external_urls: ExternalUrls
    uri: str


@dataclass_json
@dataclass
class Actions:
    interrupting_playback: bool
    pausing: bool
    resuming: bool
    seeking: bool
    skipping_next: bool
    skipping_prev: bool
    toggling_repeat_context: bool
    toggling_shuffle: bool
    toggling_repeat_track: bool
    transferring_playback: bool


@dataclass_json
@dataclass
class PlaybackState:
    device: Device
    repeat_state: str
    shuffle_state: bool
    context: Context
    timestamp: int
    progress_ms: int
    is_playing: bool
    item: Track
    currently_playing_type: str
    actions: Actions



from typing import List, Union
from shrillecho.api.api_client import SpotifyClient
from shrillecho.types.artist_types import Artist, FollowedArtists
from shrillecho.types.playlist_types import SimplifiedPlaylistObject
from shrillecho.types.track_types import Track
from shrillecho.types.user_types import BaseUserProfile, CurrentUserProfile, UserProfile


class UserController:

    def __init__(
        self, sp: SpotifyClient, user: Union[UserProfile | CurrentUserProfile]
    ):
        self._user: Union[UserProfile | CurrentUserProfile] = user
        self._sp = sp

    async def get_followed_artists(self) -> List[Artist]:
        artists: List[Artist] = []
        followed_artist_page: FollowedArtists = (
            await self._sp.follow.get_followed_artists(limit=50, after=None)
        )
        while len(followed_artist_page.artists.items) != 0:
            after = followed_artist_page.artists.items[-1].id
            artists.extend(followed_artist_page.artists.items)
            followed_artist_page: FollowedArtists = (
                await self._sp.follow.get_followed_artists(limit=50, after=after)
            )
        return artists

    async def fetch_all_user_public_tracks(self) -> List[Track]:
        tracks: List[Track] = []
        playlists: List[SimplifiedPlaylistObject] = (
            await self._sp.playlists.user_playlists_batch(user=self._user.id)
        )
        for playlist in playlists:
            tracks.extend(
                await self._sp.playlists.tracks_batch(playlist_id=playlist.id)
            )
        return tracks



from __future__ import annotations
import json
from typing import List
import requests
from shrillecho.api.api_client import SpotifyClient
from shrillecho.api.task_scheduler import TaskScheduler
from shrillecho.utility.track_utils import TrackUtils
from shrillecho.types.album_types import SimplifiedAlbum, SimplifiedTrack
from shrillecho.types.artist_types import Artist, FollowerCount
from shrillecho.types.track_types import Track
from shrillecho.utility.cache import cached
import logging

# logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ArtistController:

    def __init__(self, sp: SpotifyClient, artist: Artist):
        self._artist = artist
        self._sp = sp

    @staticmethod
    async def create_from_id(sp: SpotifyClient, artist_id: str = None):
        artist: Artist = await sp.artists.get(artist_id=artist_id)
        return ArtistController(sp=sp, artist=artist)

    async def related(self) -> List[Artist]:
        return await self._sp.artists.related(artist_id=self._artist.id)

    async def albums(self) -> List[SimplifiedAlbum]:
        return await self._sp.artists.artist_albums_batch(artist_id=self._artist.id)

    async def get_all_tracks(self) -> List[SimplifiedTrack]:
        artist_albums: List[SimplifiedAlbum] = await self.albums()
        task_scheduluer = TaskScheduler()
        tasks = [
            (self._sp.albums.album_tracks_batch, (album.id,)) for album in artist_albums
        ]
        task_scheduluer.load_tasks(tasks)
        return await task_scheduluer.execute_tasks(batch_size=1)

    async def get_artist_track_ids(self) -> List[str]:
        """Return all tracks in the form of there ids

        Returns:
            List[str]: all artist tracks in for of ids
        """
        return [item.id for item in await self.get_all_tracks()]

    async def get_artist_tracks(self, unique: bool = True) -> List[Track]:
        all_ids = await self.get_artist_track_ids()
        tracks: List[Track] = await TrackUtils.fetch_several_tracks(
            sp=self._sp, track_ids=all_ids
        )
        if unique:
            return set(tracks)

        return tracks

    """
    TODO: New releaseas
    """
    """
     @staticmethod
    async def get_artist_albums(sp: SpotifyClient, artist_id: str, simple=False) -> List[Album] | List[SimplifiedAlbum]:
        
        artist_albums: List[SimplifiedAlbum] = await sp.artist_albums(artist=artist_id)

        if simple:
            return artist_albums

        album_ids = [album.id for album in artist_albums]
        
        return AlbumUtils.fetch_several_albums(sp, album_ids=album_ids)
    
    @staticmethod
    async def get_artist_new_releases(sp: SpotifyClient, artist_id: str, earliest_date: str ='2024-03-01' ) -> List[SimplifiedAlbum]:
       
        albums: List[SimplifiedAlbum] = []
        artist_albums: ArtistAlbums = await sp.artist_albums_single(artist=artist_id, offset=0)
        albums.extend(artist_albums.items)
        next = artist_albums.next
        
        while next != None:
           

            offset: int = artist_albums.offset + 50
            artist_albums: ArtistAlbums = await sp.artist_albums_single(artist=artist_id, offset=offset)

            if len(artist_albums.items) == 0:
                break

            last_album = artist_albums.items[-1]
            next = artist_albums.next

            if is_earlier(last_album.release_date, earliest_date):
                break

            albums.extend(artist_albums.items)

         
        return albums
    """

    """
    TODO: obscure artists
    
    """


# @staticmethod
# async def most_obscure_artists(sp: SpotifyClient, artist: str, artists: List[Artist], depth: int = 0) -> List[Artist]:

#     if depth == 10:
#         return artists

#     artist_id = get_id("artist", artist)

#     related_artists: List[Artist] = await sp.artist_related(artist=artist_id)

#     most_obscure = math.inf
#     most_obscure_artist = None

#     for artist in related_artists:
#         followers: FollowerCount = await SpotifyArtistUtil.followers(sp, artist=artist.id)
#         follower_count = followers.followers

#         if follower_count < most_obscure and artist not in artists:
#             most_obscure = follower_count
#             most_obscure_artist = artist

#     artists.append(most_obscure_artist)
#     depth += 1

#     return await SpotifyArtistUtil.most_obscure_artists(sp, most_obscure_artist.id, artists, depth)



BASE_API_URL = "https://api.spotify.com/v1"


class SPOTIFY_URLS:

    class TRACKS:
        """
        URLs concerning tracks
        """

        ANALYZE: str = BASE_API_URL + "/audio-analysis/{id}"
        FEATURES: str = BASE_API_URL + "/audio-features/{id}"
        MULTI_FEATURES: str = BASE_API_URL + "/audio-features"
        SEVERAL: str = BASE_API_URL + "/tracks"
        ONE: str = BASE_API_URL + "/tracks/{id}"
        RECOMMENDATIONS: str = BASE_API_URL + "/recommendations"

    class PLAYLIST:
        """
        URLs concerning the playlists
        """

        ONE: str = BASE_API_URL + "/playlists/{playlist_id}"
        CREATE: str = BASE_API_URL + "/users/{user_id}/playlists"
        MY_PLAYLISTS: str = BASE_API_URL + "/me/playlists"
        USER: str = BASE_API_URL + "/users/{user_id}/playlists"
        COVER: str = BASE_API_URL + "/playlists/{playlist_id}/images"
        TRACKS: str = BASE_API_URL + "/playlists/{playlist_id}/tracks"

    class SEARCH:
        """
        URLs concerning the playlists
        """

        ITEM: str = BASE_API_URL + "/search"

    class LIBRARY:
        """
        URLs concerning the user library
        """

        CONTAINS_ALBUM: str = BASE_API_URL + "/me/albums/contains"
        CONTAINS_TRACK: str = BASE_API_URL + "/me/tracks/contains"
        CONTAINS_SHOWS: str = BASE_API_URL + "/me/shows/contains"

        ALBUMS: str = BASE_API_URL + "/me/albums"
        TRACKS: str = BASE_API_URL + "/me/tracks"
        SHOWS: str = BASE_API_URL + "/me/shows"

    class ALBUM:
        """
        URLs concerning the album
        """

        ONE: str = BASE_API_URL + "/albums/{id}"
        TRACKS: str = BASE_API_URL + "/albums/{id}/tracks"
        MULTIPLE: str = BASE_API_URL + "/albums"

    class ARTIST:
        """
        URLs concerning the artist
        """

        ONE: str = BASE_API_URL + "/artists/{id}"
        ALBUM: str = BASE_API_URL + "/artists/{id}/albums"
        TOP_TRACKS: str = BASE_API_URL + "/artists/{id}/top-tracks"
        SIMILAR_ARTISTS: str = BASE_API_URL + "/artists/{id}/related-artists"
        SEVERAL: str = BASE_API_URL + "/artists/"

    class USER:
        """
        URLs concerning a spotify user profile
        """

        ME: str = BASE_API_URL + "/me"
        USER: str = BASE_API_URL + "/users/{user_id}"

    class FOLLOW:
        """
        URLs concerning the follow endpoint
        """

        CONTAINS: str = BASE_API_URL + "/me/following/contains"
        CONTAINS_PLAYLIST: str = (
            BASE_API_URL + "/playlists/{playlist_id}/followers/contains"
        )
        HUMAN: str = BASE_API_URL + "/me/following"
        PLAYLIST: str = BASE_API_URL + "/playlists/{playlist_id}/followers"
        FOLLOWING_ARTISTS: str = BASE_API_URL + "/me/following?type=artist"

    class PLAYER:
        """
        URLs concerning the player
        """

        PLAYER: str = BASE_API_URL + "/me/player"
        QUEUE: str = BASE_API_URL + "/me/player/queue"
        DEVICES: str = BASE_API_URL + "/me/player/devices"
        RECENTLY: str = BASE_API_URL + "/me/player/recently-played"
        PLAYING: str = BASE_API_URL + "/me/player/currently-playing"

        PAUSE: str = BASE_API_URL + "/me/player/pause"
        SEEK: str = BASE_API_URL + "/me/player/seek"
        REPEAT: str = BASE_API_URL + "/me/player/repeat"
        VOLUME: str = BASE_API_URL + "/me/player/volume"
        NEXT: str = BASE_API_URL + "/me/player/next"
        PREVIOUS: str = BASE_API_URL + "/me/player/previous"
        PLAY: str = BASE_API_URL + "/me/player/play"
        SHUFFLE: str = BASE_API_URL + "/me/player/shuffle"
